package csbi

import (
	"bufio"
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"os/exec"
	"path/filepath"
	"time"

	spb "code.fbi.h-da.de/danet/api/go/gosdn/southbound"

	"github.com/docker/docker/api/types"
	"github.com/docker/docker/client"
	"github.com/docker/docker/pkg/archive"
	"github.com/google/uuid"
	"github.com/prometheus/client_golang/prometheus"
	log "github.com/sirupsen/logrus"
)

// nolint
type ErrorLine struct {
	Error       string      `json:"error"`
	ErrorDetail ErrorDetail `json:"errorDetail"`
}

// nolint
type ErrorDetail struct {
	Message string `json:"message"`
}

func buildImage(d Deployment, dockerClient *client.Client) error {
	labels := prometheus.Labels{"type": spb.Type_CONTAINERISED.String()}
	start := promStartHook(labels, buildsTotal)

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*300)
	defer cancel()

	p := filepath.Join(d.ID.String(), "/")
	tar, err := archive.TarWithOptions(p, &archive.TarOptions{})
	if err != nil {
		return err
	}

	opts := types.ImageBuildOptions{
		Dockerfile: "Dockerfile",
		Tags:       []string{d.Name},
		Remove:     true,
	}
	res, err := dockerClient.ImageBuild(ctx, tar, opts)
	if err != nil {
		return err
	}
	defer res.Body.Close()

	err = print(res.Body)
	if err != nil {
		return err
	}
	promEndHook(labels, start, buildDurationSecondsTotal, buildDurationSeconds)
	return nil
}

func print(rd io.Reader) error {
	var lastLine string

	scanner := bufio.NewScanner(rd)
	for scanner.Scan() {
		lastLine = scanner.Text()
		fmt.Println(scanner.Text())
	}

	errLine := &ErrorLine{}
	json.Unmarshal([]byte(lastLine), errLine)
	if errLine.Error != "" {
		return errors.New(errLine.Error)
	}

	return scanner.Err()
}

// buildPlugin builds a go plugin from ygot generated go code provided within
// a plugin folder.
func buildPlugin(id uuid.UUID) error {
	labels := prometheus.Labels{"type": spb.Type_PLUGIN.String()}
	start := promStartHook(labels, buildsTotal)

	var stderr bytes.Buffer
	buildDir := id.String()

	if err := executeGoCommand(buildDir, &stderr, []string{"go", "mod", "tidy"}); err != nil {
		log.Error(err)
		return err
	}

	stderr.Reset()

	buildCommand := []string{
		"go",
		"build",
		"-o",
		"./plugin",
		"./gostructs.go",
	}
	if err := executeGoCommand(buildDir, &stderr, buildCommand); err != nil {
		log.Error(stderr.String())
		return err
	}
	promEndHook(labels, start, buildDurationSecondsTotal, buildDurationSeconds)
	return nil
}

/*
executeGoCommand runs a go command. Therefore it creates a new *exec.Cmd and
adds the provided build directory as string, a byte buffer and the build
commands as string slice.

Example for a build command slice:

buildCommand := []string{
    "build",
    "-o",
    outputPath,
    sourcePath,
}
*/
func executeGoCommand(dir string, stderr *bytes.Buffer, buildCommand []string) error {
	cmd := exec.Command(buildCommand[0], buildCommand[1:]...)
	cmd.Dir = dir
	cmd.Stderr = stderr

	return cmd.Run()
}
