package csbi

import (
	"html/template"

	"github.com/openconfig/ygot/ygen"
)

var pluginStruct = ygen.GoStructCodeSnippet{
	StructName: "Csbi",
	StructDef: `type Csbi struct {
	id          uuid.UUID
}`,
	Methods: `
// SetNode injects schema specific model representation to the transport.
// Needed for type assertion.
func (csbi *Csbi) SetNode(schema *yang.Entry, root interface{}, path *gpb.Path, val interface{}, opts ...ytypes.SetNodeOpt) error {
	return ytypes.SetNode(schema, root.(*Device), path, val, opts...)
}

// Unmarshal injects schema specific model representation to the transport.
// Needed for type assertion.
func (csbi *Csbi) Unmarshal(bytes []byte, path *gpb.Path, goStruct ygot.ValidatedGoStruct, opt ...ytypes.UnmarshalOpt) error {
	return unmarshal(csbi.Schema(), bytes, path, goStruct, opt...)
}

//unmarshal parses a gNMI response to a go struct.
func unmarshal(schema *ytypes.Schema, bytes []byte, path *gpb.Path, goStruct ygot.ValidatedGoStruct, opt ...ytypes.UnmarshalOpt) error {
	defer func() {
		if r := recover(); r != nil {
			log.Error(r.(error))
		}
	}()

	// Load SBI definition
	root, err := ygot.DeepCopy(schema.Root)
	if err != nil {
		return err
	}
	validatedDeepCopy, ok := root.(ygot.ValidatedGoStruct)
	if !ok {
		return &errors.ErrInvalidTypeAssertion{}
	}

	// returns the node we want to fill with the data contained in 'bytes',
	// using the specified 'path'.
	createdNode, _, err := ytypes.GetOrCreateNode(schema.RootSchema(), validatedDeepCopy, path)
	if err != nil {
		return err
	}
	validatedCreatedNode, ok := createdNode.(ygot.ValidatedGoStruct)
	if !ok {
		return &errors.ErrInvalidTypeAssertion{}
	}

	if err := Unmarshal(bytes, validatedCreatedNode, opt...); err != nil {
		return err
	}

	opts := []ygot.MergeOpt{&ygot.MergeOverwriteExistingFields{}}
	return ygot.MergeStructInto(goStruct, validatedDeepCopy, opts...)
}

func (csbi *Csbi) Schema() *ytypes.Schema {
	schema, err := Schema()
	if err != nil {
		log.Fatal(err)
	}
	return schema
}

// SetID sets the ID of the cSBI to the provided UUID
func (csbi *Csbi) SetID(id uuid.UUID) error {
    csbi.id = id
    return nil
}

// ID returns the Southbound's UUID
func (csbi *Csbi) ID() uuid.UUID {
	return csbi.id
}

// Type returns the Southbound's type
func (csbi *Csbi) Type() spb.Type {
	return spb.Type_CONTAINERISED
}

func main() {
    handshakeConfig := plugin.HandshakeConfig{
	    ProtocolVersion:  1,
	    MagicCookieKey:   "plugin_sbi",
	    MagicCookieValue: "cMBgvDYv89",
    }

    logger := hclog.New(&hclog.LoggerOptions{
		Level:      hclog.Trace,
		Output:     os.Stderr,
		JSONFormat: true,
	})

	sbi := &Csbi{}
	// pluginMap is the map of plugins we can dispense.
	var pluginMap = map[string]plugin.Plugin{
        "plugin_sbi": &sbp.SouthboundPlugin{Impl: sbi},
	}

	logger.Debug("message from plugin", "foo", "bar")

	plugin.Serve(&plugin.ServeConfig{
		HandshakeConfig: handshakeConfig,
		Plugins:         pluginMap,
	})
}`,
}

const pluginImportAmendmend = `
    "os"

    "code.fbi.h-da.de/danet/gosdn/nucleus/errors"
	spb "code.fbi.h-da.de/danet/api/go/gosdn/southbound"
    sbp "code.fbi.h-da.de/danet/gosdn/nucleus/plugins/southbound"


	"github.com/hashicorp/go-hclog"
	"github.com/hashicorp/go-plugin"
	"github.com/google/uuid"

	log "github.com/sirupsen/logrus"
)
`

var templater *template.Template

const pluginGoModTemplate = `module << .Module.Path >>

go << .GoVersion >>

replace "code.fbi.h-da.de/danet/gosdn" => ../../gosdn

require (<<range $element := .Dependencies>>
	<<$element.Path>> <<$element.Version>>
<<end>>)
`

type Module struct {
	Path string `json:"Path"`
}

type Dependency struct {
	Path    string `json:"Path"`
	Version string `json:"Version"`
	//Indirect bool   `json:"Indirect,omitempty"`
}

// GoMod represents a go.mod file used for templates.
type GoMod struct {
	Module       Module       `json:"Module"`
	GoVersion    string       `json:"Go"`
	Dependencies []Dependency `json:"Require"`
}
